// Copyright 2015 Google Inc. All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef FILE_CACHE_H_
#define FILE_CACHE_H_

#include <string>
#include <unordered_set>

#include "file.h"

class MakefileCacheManager {
 public:
  const Makefile& ReadMakefile(const std::string& filename);
  void GetAllFilenames(std::unordered_set<std::string>* out);
  void AddExtraFileDep(std::string_view dep);

  static MakefileCacheManager& Get();

 private:
  MakefileCacheManager() = default;
  MakefileCacheManager(const MakefileCacheManager&) = delete;
  MakefileCacheManager(MakefileCacheManager&&) = delete;
  std::unordered_map<std::string, Makefile> cache_;
  std::unordered_set<std::string> extra_file_deps_;
};

#endif  // FILE_CACHE_H_
