// Copyright 2015 Google Inc. All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef IO_H_
#define IO_H_

#include <stdio.h>

#include <string>
#include <string_view>

void DumpInt(FILE* fp, int v);
void DumpString(FILE* fp, std::string_view s);

int LoadInt(FILE* fp);
bool LoadString(FILE* fp, std::string* s);

struct ScopedFile {
 public:
  explicit ScopedFile(FILE* fp) : fp_(fp) {}

  ~ScopedFile() {
    if (fp_)
      fclose(fp_);
  }

 private:
  FILE* fp_;
};

#endif  // IO_H_
