#!/bin/sh
#
# Copyright 2016 Google Inc. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

set -e

mk="$@"

cat <<EOF > Makefile
CLVAR := FAIL
MFVAR := FAIL
FILEVAR := PASS
all:
	@echo \$(ENVVAR) \$(origin ENVVAR)
	@echo \$(MFVAR) \$(origin MFVAR)
	@echo \$(CLVAR) \$(origin CLVAR)
	@echo \$(FILEVAR) \$(origin FILEVAR)
EOF

export ENVVAR=PASS
export FILEVAR=FAIL
export MAKEFLAGS="MFVAR=PASS CLVAR=FAIL"

${mk} CLVAR=PASS 2> /dev/null
