#!/bin/bash
#
# Copyright 2021 Google Inc. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This just tests that kati doesn't crash when using --empty_ninja_file

set -u

mk="$@"

cat <<EOF > Makefile
result:
	@echo "Foo" > \$@
EOF

if echo "${mk}" | grep -qv "kati"; then
  # Make doesn't support --empty_ninja_file
  ${mk} --warn 2>&1
else
  ${mk} --ninja --empty_ninja_file --warn 2>&1
fi
